/*   changes from v4:
- Increased depth of Vitesse Spy Fifos
changes from v3:
- Added VitResetN on WriteOnlyCntrReg
- use both clocks
- TTC for the HTR inputs, RxClk for the Vitesse
- spy HTR data with TTC clock.
changes from v2:
- swapped input clocks of BUFGMUX
- added register for real time value of Vitesse flags.
- change addresses to make some commands compatible with SLB:
  START <=> write a 2 on addr 0x61    ---> Start SpyFIFOs
  STOP  <=> write a 4 on addr 0x61    ---> Stop  SpyFIFOs
  STATUS<=> read addr 0x60, 0x70      ---> read Vitesse_StatReg and Fifo_StatReg
  CSTAT <=> read from 0x0 or 0x10     ---> read CntrReg or StatusReg
  REGS  <=> read 0x60 and either 0x62-6F or 0x72-7F  ---> unused in SLBS

TBD 
+ Add a bit: "Enable Spy On Valid Data" EnSOVD on CntrReg
*/
/* Note that all 120 Mb/s inputs from the Vitesse are synchronized with
REFCLK. The inputs are valid and stable on rising edges of REFCLK 
other than a round-trip propagation delay (? ns) */

`include "define.v"
module slbsTop 
 ( 
 input  CLKPLUS, CLKMINUS, // 40MHz differential RX_CLK
 input  BCOPLUS, BCOMINUS, // RX_BC0 common to all SLB boards
 input  TTC_CK40Des1, TTC_BRCST_STRB1, TTC_BCR,
 input  LOC_C_S,  LOC_R_W, 
 input  ErrA, ErrB, ErrC, ErrD,        //  120 Mb/s inputs
        IdleA, IdleB, IdleC, IdleD,    //  120 Mb/s inputs
        KChA, KChB, KChC, KChD,        //  120 Mb/s inputs
 input  [7:0] RecA, RecB, RecC, RecD,  //  120 Mb/s inputs
 input  [7:2] TTC_BRCST,      
 input  [8:0] SYNC1, SYNC2, SYNC3, SYNC4, 
              SYNC5, SYNC6, SYNC7, SYNC8,   // Data In
 input  [`LB_Addr_Size-1:0]  LOCADD,  // The Address bus.
 inout  [`REG_WIDTH-1:0]     LOCBUS,  // The bidir Data bus 

 output TPMINUS, TPPLUS, // to RJ45 J8
 output TRSTN, 
        REFCLK,     // to MC100EPT22 and then to Vitesse mezzanine
        PECL_OUT,  // to two MC100EPT22 and then connectors
        VitResetN          );

 wire   
  TTC_Hard_rst, TTC_Soft_rst, TTC_Start, TTC_Stop, TTC_AsyncFifoRst,
  VME_Hard_rst, VME_Soft_rst, VME_Start, VME_Stop, VME_AsyncFifoRst,
  run_status,   BunchEr, GateTTCcommands, Hard_rst, Soft_rst;
 wire [7:0] LastTTCcommand;
 wire [15:0] TTC_test0, TTC_test1, TTC_test2, TTC_test3,
             TTC_test4, TTC_test5, TTC_test6, TTC_test7,
             TTC_test8, TTC_test9, TTC_testA, TTC_testB,
             TTC_testC, TTC_testD, TTC_testE, TTC_testF;
 wire [`BCN_WIDTH-1:0] MaxBCN; 
 wire [8:1]  TPSpyFifoEmpty,  TPSpyFifoFull,  TPSpyFifoRd;
 wire [4:1]  VitSpyFifoEmpty, VitSpyFifoFull, VitSpyFifoRd;     
 wire [8:0]  TPSpyFIFO1_out, TPSpyFIFO2_out, TPSpyFIFO3_out, TPSpyFIFO4_out,
             TPSpyFIFO5_out, TPSpyFIFO6_out, TPSpyFIFO7_out, TPSpyFIFO8_out;
 wire [10:0] VitSpyFIFO1_out, VitSpyFIFO2_out,
             VitSpyFIFO3_out, VitSpyFIFO4_out;	
 wire [6:0]  pad7_1,  pad7_2,  pad7_3,  pad7_4;
 wire [26:0] pad27_1, pad27_2, pad27_3, pad27_4,
             pad27_5, pad27_6, pad27_7, pad27_8;

//  ********************************************************************
//      ********************  CLOCK SECTION    *****************
//            *********************************************

// XST was stupid at handling clocks and clock buffer, so we instantiate
// all the buffers and inihbit XST from changing them with the attribute.
 wire   RxClk, clk, clk2x, clk0, clk2x_dll;
 wire   [7:0] DLL_status;   // DLL_status[1] asserted if .CLKIN stopped
 wire   DLL_lock, ClkFx, ClkFx180, clk3x, clk3x180;
 reg    ResetDLL, DLL_lock_D;
 reg    [14:0] RstCnt; // counter for proper reset of DLL 
 reg    [2:0] UnlockCnt; // counter for proper reset of DLL 

 // Instantiate the Differential Global Input Buffer 
  IBUFGDS_LVPECL_25 difclkbuf (.I(CLKPLUS), .IB(CLKMINUS), .O(RxClk));
  IBUFG ttcclkbuf ( .I(TTC_CK40Des1), .O(TTC_CK) );

// Generate a DLL reset (3-ticks) after FPGA is configured, see:
// www.xilinx.com/publications/xcellonline/xcell_51/xc_particle51.htm
  FDS flop1 (.D(1'b0), .C(RxClk), .Q(out1), .S(1'b0)); 
  FD  flop2 (.D(out1), .C(RxClk), .Q(out2)); 
  FD  flop3 (.D(out2), .C(RxClk), .Q(out3)); 
  FD  flop4 (.D(out3), .C(RxClk), .Q(out4)); 
  assign ConfigRst = (out2 | out3 | out4); 

  DCM MainDCM
   (.CLKIN(RxClk), .CLKFB(clk),  .RST(ConfigRst),  
    .CLK0(clk0),   .CLK90(),  .CLK180(),  .CLK270(),
    .CLK2X(clk2x_dll),  .CLK2X180(),
    .CLKDV(),   .CLKFX(ClkFx),   .CLKFX180(ClkFx180),
    .LOCKED(DLL_lock),  .STATUS(DLL_status),
    .DSSEN(1'b0),
    .PSCLK(1'b0),  .PSEN(1'b0),  .PSINCDEC(1'b0), .PSDONE() );

  //synthesis attribute DFS_FREQUENCY_MODE of MainDCM is "LOW" 
  //synthesis attribute CLKFX_DIVIDE of MainDCM is "1" 
  //synthesis attribute CLKFX_MULTIPLY of MainDCM is "3" 
  //synthesis attribute STARTUP_WAIT of MainDCM is "FALSE" 

/* If tests show that DCM looses lock, consider to
   increase the FACTORY_JF setting to 'FFFF  (Xilinx Answer 13756).
   Anyway lab test in Dec03 showed that the DLL relock spontaneously
   when the TTCrx was loosing lock and then relocking   */
/* If there is a chance that RxClk stops completly, the LOCKED 
   signal will not update; consider the STATUS bit "CLK_IN_STOPPED".
   I do not implement it as I don't expect this occurence, although
   trivial mechanical troubles of the TTC-PMC mezzanine could do it. 
   Note that if the TTCrx looses lock, TTC_clk will keep toggling at 
   the TTCrx natural oscillation frequency, but will not stop.    */

// ResetDLL has some "feedback" problems when connected to DCM,
// so far I don't use it.
 always @ (posedge TTC_CK) begin  //TTC_CK should toggle even if no cable 
    if (DLL_lock == 0) begin
         RstCnt <= RstCnt + 1;
         if (RstCnt == 15'h 7FFF)
               ResetDLL <= 1;  // reset no more then every 2^15 25ns
         else     ResetDLL <= 0;
    end
    DLL_lock_D <= DLL_lock;
    if (Hard_rst)  UnlockCnt <= 0;
    else if (DLL_lock_D & (~DLL_lock) & (UnlockCnt != 3'b111)) //hi->lo transition
             UnlockCnt <= UnlockCnt + 1;   // increment (stop at max count)
 end //always 

 BUFG CK40BUF (.I(clk0), .O(clk));  // Use clk as internal clock 
 BUFG CK2XBUF (.I(clk2x_dll), .O(clk2x)); 
 BUFG CK3XBUF    (.I(ClkFx),    .O(clk3x));
 BUFG CK3X180BUF (.I(ClkFx180), .O(clk3x180));

 assign REFCLK = clk3x; // to U2A and then to Wisc_Mezzanine.

//       *************  END OF CLOCK SECTION  *****************
//  *****************************************************************

 IBUFDS_LVPECL_25 RxBC0buf (.I(BCOPLUS), .IB(BCOMINUS), .O(RxBC0));
 assign PECL_OUT = RxBC0;
 assign TRSTN = 1'b1; // Wisc. Rec_Mezz JTAG reset.
 OBUFDS_LVPECL_25 RJ45pair (.I(clk3x), .O(TPPLUS), .OB(TPMINUS));

// **** 4 Vitesse Spy FIFOs (1 per Vitesse channel) ***
// 2 BRAM per channel : 2047-deep fifos

FifoIc2047x18  VitSpyFIFO1 
 ( .fifo_gsr_in(Hard_rst),          .write_data_in({7'b0, ErrA, IdleA, KChA, RecA}), 
   .write_clock_in(clk3x),          .write_enable_in(run_status),
   .read_clock_in(clk),             .read_enable_in(VitSpyFifoRd[1]),
   .read_data_out({pad7_1, VitSpyFIFO1_out}),  .fifostatus_out(),
   .full_out(VitSpyFifoFull[1]),    .empty_out(VitSpyFifoEmpty[1])  );

FifoIc2047x18  VitSpyFIFO2 
 ( .fifo_gsr_in(Hard_rst),          .write_data_in({7'b0, ErrB, IdleB, KChB, RecB}), 
   .write_clock_in(clk3x),          .write_enable_in(run_status),
   .read_clock_in(clk),             .read_enable_in(VitSpyFifoRd[2]),
   .read_data_out({pad7_2, VitSpyFIFO2_out}),  .fifostatus_out(),
   .full_out(VitSpyFifoFull[2]),    .empty_out(VitSpyFifoEmpty[2])  );

FifoIc2047x18  VitSpyFIFO3 
 ( .fifo_gsr_in(Hard_rst),          .write_data_in({7'b0, ErrC, IdleC, KChC, RecC}), 
   .write_clock_in(clk3x),          .write_enable_in(run_status),
   .read_clock_in(clk),             .read_enable_in(VitSpyFifoRd[3]),
   .read_data_out({pad7_3, VitSpyFIFO3_out}),  .fifostatus_out(),
   .full_out(VitSpyFifoFull[3]),    .empty_out(VitSpyFifoEmpty[3])  );

FifoIc2047x18  VitSpyFIFO4 
 ( .fifo_gsr_in(Hard_rst),          .write_data_in({7'b0, ErrD, IdleD, KChD, RecD}), 
   .write_clock_in(clk3x),          .write_enable_in(run_status),
   .read_clock_in(clk),             .read_enable_in(VitSpyFifoRd[4]),
   .read_data_out({pad7_4, VitSpyFIFO4_out}),  .fifostatus_out(),
   .full_out(VitSpyFifoFull[4]),    .empty_out(VitSpyFifoEmpty[4])  );

// **** 8 TP Spy FIFOs (1 per Trigger Primitive) ***
// 1 BRAM per fifo (511x36, some waste of RAM here)

fifo_ic_v2  TPSpyFIFO1 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC1}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[1]),
   .read_data_out({pad27_1, TPSpyFIFO1_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[1]),      .empty_out(TPSpyFifoEmpty[1])  );

fifo_ic_v2  TPSpyFIFO2 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC2}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[2]),
   .read_data_out({pad27_2, TPSpyFIFO2_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[2]),      .empty_out(TPSpyFifoEmpty[2])  );

fifo_ic_v2  TPSpyFIFO3 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC3}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[3]),
   .read_data_out({pad27_3, TPSpyFIFO3_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[3]),      .empty_out(TPSpyFifoEmpty[3])  );

fifo_ic_v2  TPSpyFIFO4 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC4}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[4]),
   .read_data_out({pad27_4, TPSpyFIFO4_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[4]),      .empty_out(TPSpyFifoEmpty[4])  );

fifo_ic_v2  TPSpyFIFO5 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC5}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[5]),
   .read_data_out({pad27_5, TPSpyFIFO5_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[5]),      .empty_out(TPSpyFifoEmpty[5])  );

fifo_ic_v2  TPSpyFIFO6 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC6}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[6]),
   .read_data_out({pad27_6, TPSpyFIFO6_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[6]),      .empty_out(TPSpyFifoEmpty[6])  );

fifo_ic_v2  TPSpyFIFO7 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC7}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[7]),
   .read_data_out({pad27_7, TPSpyFIFO7_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[7]),      .empty_out(TPSpyFifoEmpty[7])  );

fifo_ic_v2  TPSpyFIFO8 
 ( .fifo_gsr_in(Hard_rst),            .write_data_in({27'b0, SYNC8}), 
   .write_clock_in(TTC_CK),           .write_enable_in(run_status),
   .read_clock_in(clk),               .read_enable_in(TPSpyFifoRd[8]),
   .read_data_out({pad27_8, TPSpyFIFO8_out}), .fifostatus_out(),
   .full_out(TPSpyFifoFull[8]),      .empty_out(TPSpyFifoEmpty[8])  );


////////////////////////////////////////////////////////

LocBusSlave LocBusSlave (
    .clk(clk),           .rst(Hard_rst), 
    .LocAddr(LOCADD),    .LocData(LOCBUS), 
    .LocCS_b(LOC_C_S),   .LocWRITE_b(LOC_R_W), 
    .LB_STATE(),    .Din(),  .Dout(),      .Tri(), 
    .DLL_lock(DLL_lock),    .UnlockCnt(UnlockCnt),
    .ErrA(ErrA), .IdleA(IdleA), .KChA(KChA), 
    .ErrB(ErrB), .IdleB(IdleB), .KChB(KChB),
    .ErrC(ErrC), .IdleC(IdleC), .KChC(KChC),
    .ErrD(ErrD), .IdleD(IdleD), .KChD(KChD),
    .LastTTCcommand(LastTTCcommand), .GateTTCcommands(GateTTCcommands), 
    .VME_Hard_rst(VME_Hard_rst),     .VME_Soft_rst(VME_Soft_rst), 
    .VME_Start(VME_Start),           .VME_Stop(VME_Stop), 
    .VME_VitResetN(VitResetN),       .VME_AsyncFifoRst(VME_AsyncFifoRst), 
    .run_status(run_status),    
    .SlbsID(),      .ChMask(),
    .TPSpyFifoRd(TPSpyFifoRd),
    .TPSpyFifoEmpty(TPSpyFifoEmpty),   .TPSpyFifoFull(TPSpyFifoFull), 
    .TPSpyFIFO1_out(TPSpyFIFO1_out),   .TPSpyFIFO2_out(TPSpyFIFO2_out),
    .TPSpyFIFO3_out(TPSpyFIFO3_out),   .TPSpyFIFO4_out(TPSpyFIFO4_out),
    .TPSpyFIFO5_out(TPSpyFIFO5_out),   .TPSpyFIFO6_out(TPSpyFIFO6_out),
    .TPSpyFIFO7_out(TPSpyFIFO7_out),   .TPSpyFIFO8_out(TPSpyFIFO8_out),
    .VitSpyFifoEmpty(VitSpyFifoEmpty), .VitSpyFifoFull(VitSpyFifoFull),
    .VitSpyFifoRd(VitSpyFifoRd),      
    .VitSpyFIFO1_out(VitSpyFIFO1_out), .VitSpyFIFO2_out(VitSpyFIFO2_out),
    .VitSpyFIFO3_out(VitSpyFIFO3_out), .VitSpyFIFO4_out(VitSpyFIFO4_out),
    .TTC_test0(TTC_test0),       .TTC_test1(TTC_test1), 
    .TTC_test2(TTC_test2),       .TTC_test3(TTC_test3), 
    .TTC_test4(TTC_test4),       .TTC_test5(TTC_test5), 
    .TTC_test6(TTC_test6),       .TTC_test7(TTC_test7), 
    .TTC_test8(TTC_test8),       .TTC_test9(TTC_test9), 
    .TTC_testA(TTC_testA),       .TTC_testB(TTC_testB), 
    .TTC_testC(TTC_testC),       .TTC_testD(TTC_testD), 
    .TTC_testE(TTC_testE),       .TTC_testF(TTC_testF), 
    .MaxBCN(MaxBCN),             .BunchEr(BunchEr), 
    .CounterModeSel()    );

  TTC_BrcstDecoder TTC_BrcstDecoder (
   .clk(clk),      .clk2x(clk2x),       .reset(VME_Hard_rst),
   .tTTC_BCntRes(TTC_BCR),              .TTC_Brcst(TTC_BRCST),
   .TTC_BrcstStr(TTC_BRCST_STRB1),      .TTC_CK40Des1(TTC_CK),  
   .LastTTCcommand(LastTTCcommand),     .TTC_Resync(),
   .TTC_Hard_rst(TTC_Hard_rst),         .TTC_Soft_rst(TTC_Soft_rst),  
   .TTC_Start(TTC_Start),               .TTC_Stop(TTC_Stop),
   .TTC_BC0(TTC_BC0),                  
   .TTC_AsyncFifoRst(TTC_AsyncFifoRst), .TTC_TestEnable(TTC_TestEnable),
   .TTC_test0(TTC_test0), .TTC_test1(TTC_test1), .TTC_test2(TTC_test2), 
   .TTC_test3(TTC_test3), .TTC_test4(TTC_test4), .TTC_test5(TTC_test5),
   .TTC_test6(TTC_test6), .TTC_test7(TTC_test7), .TTC_test8(TTC_test8),
   .TTC_test9(TTC_test9), .TTC_testA(TTC_testA), .TTC_testB(TTC_testB),
   .TTC_testC(TTC_testC), .TTC_testD(TTC_testD), .TTC_testE(TTC_testE),
   .TTC_testF(TTC_testF),   .MaxBCN(MaxBCN),   .BunchEr(BunchEr),
   .BCN(),   .OrN() );

 commander commander (  
  .clk(clk),    .GateTTCcommands(GateTTCcommands),
  .TTC_Hard_rst(TTC_Hard_rst), .TTC_Soft_rst(TTC_Soft_rst),
  .TTC_Start(TTC_Start), .TTC_Stop(TTC_Stop), 
  .TTC_AsyncFifoRst(TTC_AsyncFifoRst),
  .VME_Hard_rst(VME_Hard_rst), .VME_Soft_rst(VME_Soft_rst), 
  .VME_Start(VME_Start), .VME_Stop(VME_Stop), .VME_AsyncFifoRst(VME_AsyncFifoRst),
  .Hard_rst(Hard_rst),  .Soft_rst(Soft_rst), 
  .AsyncFifoRst(AsyncFifoRst),
  .run_status(run_status)       );

endmodule